package vsfactory

import (
	"net"

	"github.com/rancher/wrangler/pkg/name"
	soloapiv1 "github.com/solo-io/gloo/projects/gateway/pkg/api/v1"
	solov1 "github.com/solo-io/gloo/projects/gateway/pkg/api/v1/kube/apis/gateway.solo.io/v1"
	gloov1 "github.com/solo-io/gloo/projects/gloo/pkg/api/v1"
	v1 "github.com/solo-io/gloo/projects/gloo/pkg/api/v1"
	"github.com/solo-io/gloo/projects/gloo/pkg/api/v1/core/matchers"
	"github.com/solo-io/gloo/projects/gloo/pkg/api/v1/options/headers"
	"github.com/solo-io/gloo/projects/gloo/pkg/api/v1/options/retries"
	"github.com/solo-io/solo-kit/pkg/api/v1/resources/core"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

func newGlooVirtualService(namespace, name string, hosts []string, targets ...target) *solov1.VirtualService {
	vs := &solov1.VirtualService{
		ObjectMeta: metav1.ObjectMeta{
			Name:      name,
			Namespace: namespace,
		},
		Spec: soloapiv1.VirtualService{
			VirtualHost: &soloapiv1.VirtualHost{
				Domains: hosts,
				Routes: []*soloapiv1.Route{
					{
						Matchers: []*matchers.Matcher{
							{
								PathSpecifier: &matchers.Matcher_Prefix{
									Prefix: "/",
								},
							},
						},
					},
				},
			},
		},
	}

	vs.Spec.VirtualHost.Routes[0].Action = newRouteAction(targets...)
	vs.Spec.VirtualHost.Routes[0].Options = newRoutePlugin(targets...)

	return vs
}

func tlsCopy(hostname, tlsNamespace, tlsName string, vs *solov1.VirtualService) *solov1.VirtualService {
	vsTLS := vs.DeepCopy()
	vsTLS.Name = name.SafeConcatName(vsTLS.Name, "tls", tlsName)
	host, _, err := net.SplitHostPort(hostname)
	if err != nil {
		host = hostname
	}
	vsTLS.Spec.SslConfig = &gloov1.SslConfig{
		SslSecrets: &gloov1.SslConfig_SecretRef{
			SecretRef: &core.ResourceRef{
				Name:      tlsName,
				Namespace: tlsNamespace,
			},
		},
		SniDomains: []string{
			host,
		},
	}
	vsTLS.Spec.VirtualHost.Domains = []string{hostname}
	return vsTLS
}

func newRouteAction(targets ...target) *soloapiv1.Route_RouteAction {
	if len(targets) == 0 {
		return nil
	}

	if len(targets) == 1 {
		return single(targets[0])
	}

	var dest []*gloov1.WeightedDestination

	for _, target := range targets {
		dest = append(dest, &gloov1.WeightedDestination{
			Destination: destination(target),
			Weight:      uint32(target.Weight),
		})
	}

	return &soloapiv1.Route_RouteAction{
		RouteAction: &gloov1.RouteAction{
			Destination: &gloov1.RouteAction_Multi{
				Multi: &gloov1.MultiDestination{
					Destinations: dest,
				},
			},
		},
	}
}

func destination(target target) *gloov1.Destination {
	return &gloov1.Destination{
		DestinationType: &gloov1.Destination_Kube{
			Kube: &gloov1.KubernetesServiceDestination{
				Ref: core.ResourceRef{
					Name:      target.Name,
					Namespace: target.Namespace,
				},
				Port: uint32(target.Port),
			},
		},
	}
}

func single(target target) *soloapiv1.Route_RouteAction {
	return &soloapiv1.Route_RouteAction{
		RouteAction: &gloov1.RouteAction{
			Destination: &gloov1.RouteAction_Single{
				Single: destination(target),
			},
		},
	}
}

func newRoutePlugin(targets ...target) *v1.RouteOptions {
	for _, t := range targets {
		// use envoy to retry the connection if the service is being scaled up and increase envoy timeout
		if t.ScaleIsZero {
			return &v1.RouteOptions{
				Timeout: &overallTimeout,
				Retries: &retries.RetryPolicy{
					RetryOn:       "5xx",
					NumRetries:    scaleFromZeroAttempts,
					PerTryTimeout: &perTryTimeout,
				},
				HeaderManipulation: &headers.HeaderManipulation{
					RequestHeadersToAdd: []*headers.HeaderValueOption{
						{
							Header: &headers.HeaderValue{
								Key:   rioNameHeader,
								Value: t.OriginalTarget.Name,
							},
						},
						{
							Header: &headers.HeaderValue{
								Key:   rioNamespaceHeader,
								Value: t.OriginalTarget.Namespace,
							},
						},
					},
				},
			}
		}
	}
	return nil
}
